##### Configuration

log ?= $(if $(test),,1)
profile ?= $(log)
exts ?= csv xml
test_n ?= 2
tablesSort ?= plots organisms stems

##### Vars/functions

selfDir_uZPPqC := $(dir $(lastword $(MAKEFILE_LIST)))

# Make
SHELL := /bin/bash
subMake = $(MAKE) $(@F) --directory=$(@D)
+_ = $(+:_%=)
addBeforeExt = $(basename $(2))$(1)$(suffix $(2))

# System
date = $(shell date +"%Y-%m-%d-%H-%M-%S")

# Terminal
termCols := $(shell tput cols)
esc := '['
reset := $(esc)'0m'
emph := $(esc)'7m '
endEmph := ' '$(reset)

# Commands
CP = cp -p
diff = diff --unified=2
diffVerbose = $(if $(verbose),diff --side-by-side --left-column\
--width=$(termCols),$(diff))

# Paths
root := $(selfDir_uZPPqC)..
mappings := $(root)/mappings
datasrc := $(notdir $(realpath .))
psqlOpts := --set ON_ERROR_STOP=1 --quiet
psqlAsBien := $(root)/bin/psql_vegbien $(psqlOpts)

##### General targets

all: _always maps ;

.SUFFIXES:

_always:
.PHONY: _always

clean: _always
	$(RM) $(all)

remake: _always clean all ;

%.out: %.make _always
	./$* >$@
.PRECIOUS: %.out

$(root)/%: _always
	+$(subMake)

##### SVN

svn_props: _always
	svn propset svn:ignore $$'.~*\n*.log\n*.trace\n*.out\nsrc*' .
	$(if $(wildcard maps/),svn propset svn:ignore $$'.~*' maps)
	$(if $(wildcard test/),svn propset svn:ignore $$'*.out\n*.xml' test)

##### Installation

reinstall: _always uninstall install ;

##### Maps

fullViaMap := maps/%.full.csv
allViaMaps := $(filter-out maps/VegBIEN.%.csv $(fullViaMap),\
$(wildcard maps/*.csv))
via := $(firstword $(sort $(basename $(basename $(notdir $(allViaMaps))))))

coreMap := $(mappings)/$(via)-VegBIEN.%.csv
selfMap := $(mappings)/$(via).self.%.csv

viaMaps := $(wildcard $(tablesSort:%=maps/$(via).%.csv))
viaMaps += $(filter-out $(viaMaps) $(fullViaMap),$(wildcard maps/$(via).*.csv))

autogenMaps := $(subst $(via).,VegBIEN.,$(viaMaps))
directMaps := $(autogenMaps) $(filter-out $(autogenMaps),\
$(wildcard maps/VegBIEN.*.csv))
tables := $(directMaps:maps/VegBIEN.%.csv=%)

.PRECIOUS: $(allViaMaps) $(directMaps)

maps :=

maps/VegBIEN.%.csv: maps/$(via).%.csv $(coreMap)
	$(root)/bin/join_union_sort <$+ >$@
.PRECIOUS: maps/VegBIEN.%.csv $(coreMap)

maps += $(autogenMaps)

maps/$(via).%.full.csv: maps/$(via).%.csv $(selfMap)
	env ignore=1 $(root)/bin/union <$+|$(root)/bin/sort_map >$@
maps/$(via).%.full.csv: maps/$(via).%.csv # fallback if no self map
	$(CP) $< $@
.PRECIOUS: maps/$(via).%.full.csv maps/$(via).%.csv $(selfMap)

maps += $(patsubst maps/%.csv,maps/%.full.csv,$(viaMaps))

maps: $(maps) _always ;

all += $(maps)

##### Mapping

dbFile := $(firstword $(wildcard src/db.*.sql))
inputFiles := $(wildcard src/*.csv src/*.xml)

+maps = $(filter maps/% $(mappings)/%,$(+_))
<in = $(firstword $(filter-out $(+maps),$(+_)) $(wildcard $(exts:%=src/*.$*.%)))
map = $(if $(<in),<$(<in) ,$(if $(mapEnv),env $(mapEnv) ,$(error\
No input file src/*.$*.{$(exts)} )))$(root)/map $(+maps)
map2db = env out_database=vegbien $(map)

##### Import to VegBIEN

ifneq ($(dbFile)$(inputFiles),)

log_ = $(@:-all=)$(if $(n),.n=$(n),).$(date).log
trace = $(log_:.log=.trace)
import = -(set -x; "time" env commit=1 verbose=1\
$(if $(profile),profile_to=$(trace)) $(map2db)) $(if $(log),\
$(if $(n),,>>$(log_))) 2>&1$(if $(log),$(if $(n),|tee -a $(log_)))
# don't abort on import errors, which often relate to invalid input data

import: $(addprefix import-,$(tables)) _always ;

import-%: maps/VegBIEN.%.csv _always
	$(import)
# default:
import-%: _always ;

else
import: _always ;
endif

##### Log files from import

logs := $(wildcard *.log *.trace)

rm_logs: _always
	$(RM) $(logs)

##### Verification of import

verify: $(addprefix verify-,$(tables)) _always ;

verify-%: verify.%.ref verify.%.out _always
	-$(diffVerbose) $(+_)
# don't abort on verification errors, which are expected during development
# default:
verify-%: verify.%.out _always
	$(if $(wildcard $<),cat $<)
# don't run if verify.%.out's default do-nothing action was used

verify = $(psqlAsBien) --set=datasource="'$(datasrc)'" --no-align\
--field-separator='	' --pset=footer=off --pset=null=NULL <$< >$@

verify.%.out: $(mappings)/verify.%.sql _always
	$(verify)
.PRECIOUS: verify.%.out
# default:
verify.%.out: _always ;

all += $(wildcard verify.*.out)

ifneq ($(dbFile),)
%.ref: %.ref.sql
	$(dbAsBien) $(db) <$< >$@
.PRECIOUS: %.ref
endif

##### Testing

ifneq ($(wildcard test/),)

hasOwnRef = $(filter-out %.2-step.xml,$@)
testRef = $(1:.2-step.xml=.xml).ref

define runTest
@echo "Testing $(abspath $@)..."
>$@ env test=1 n=$(test_n) $(1)
@(set -x; $(diff) $(call testRef,$@) $@) 2>&1 || { e=$$?;\
$(if $(wildcard $(call testRef,$@)),,cat $@;)\
$(if $(hasOwnRef),\
echo $(emph)"To accept new test output:"$(endEmph);\
echo "$(MAKE) $@-ok --directory=$(realpath .) --makefile=../input.Makefile";\
,\
echo $(emph)"Note: The preceding failed test is compared to another test's\
output"$(endEmph);\
echo $(emph)"When it fails, this always indicates a bug"$(endEmph);\
)\
exit $$e;}
endef

test2File = $(call runTest,$(map))

tests :=

test/$(via).%.xml: maps/$(via).%.full.csv _always
	$(test2File)
tests += test/$(via).%.xml

test/VegBIEN.%.xml: maps/VegBIEN.%.csv _always
	$(test2File)
tests += test/VegBIEN.%.xml

test/VegBIEN.%.2-step.xml: test/$(via).%.xml $(coreMap) _always
	-$(test2File)
# Don't abort tester if only 2-step test fails, as it's often finicky
tests += test/VegBIEN.%.2-step.xml

test/import.%.out: maps/VegBIEN.%.csv _always
	$(call runTest,$(map2db))
tests += test/import.%.out

.PRECIOUS: $(tests)

testOutputs := $(foreach test,$(tests),$(tables:%=$(test)))

test: _always $(testOutputs) ;

all += $(testOutputs)

# Accepts a test output: make <test_output_path>-ok
%-ok: _always
	$(CP) $* $(call testRef,$*)

else
test: _always ;
endif

##### Input-type-specific

# Each input type needs var $(mapEnv) and targets install, uninstall

ifneq ($(dbFile),)

dbEngineExt := $(subst .,,$(suffix $(basename $(notdir $(dbFile)))))
db := $(datasrc)

#### Installation

install: _always db ;

uninstall: _always rm_db ;

#### DB-engine-specific

# Each DB engine needs vars $(dbEngine), $(dbAsBien) and targets db, rm_db

ifeq ($(dbEngineExt),my)

dbEngine := MySQL

bienPassword := $(shell cat $(root)/config/bien_password)
mysqlAs = mysql --user=$(1) --password='$(bienPassword)'
mysqlAsRoot := $(call mysqlAs,root)
dbAsBien := $(call mysqlAs,bien)

dbExists = $(shell echo "SHOW DATABASES LIKE '$(db)';"|$(mysqlAsRoot))

define createDb
echo "CREATE DATABASE $(db) DEFAULT CHARACTER SET latin1;"|$(mysqlAsRoot)
-$(mysqlAsRoot) --database=$(db) <$<
endef
# ignore errors in db import so that GRANT will still be run

db: $(dbFile) _always
	$(if $(dbExists),,$(createDb))
	echo "GRANT SELECT ON $(db).* TO 'bien'@'localhost';"|$(mysqlAsRoot)

rm_db: _always
	-echo "REVOKE ALL ON $(db).* FROM 'bien'@'localhost';"|$(mysqlAsRoot)
	echo "DROP DATABASE IF EXISTS $(db);"|$(mysqlAsRoot)
# ignore errors if grant not defined

### Unrecognized DB engine

else
$(error The DB filename $(dbFile) must be db.my.sql)
endif

# Must come after dbEngine is set
mapEnv := in_engine=$(dbEngine) in_database=$(db)

#### Unrecognized input type

else
install: _always ;
uninstall: _always ;
endif
